from __future__ import division
import pandas
from IPython.core.magic import (Magics, magics_class, cell_magic)

from sasx.parse import sasx_parse
from sasx.code import sasx_preloop
from sasx.code import sasx_loop
from sasx.code import sasx_postloop


@magics_class
class SasxMagics(Magics):
    """Define Magic to run code in Simple dAta SyntaX (SASX).

      %%sasx  - Transform SASX code into Python code and execute it.
      
      Special keywords recognised by SASX :
          - data
          - set
          - drop
          - keep
          - output
          - where ?
          - _n_ ?
          - groupby ?
      
    """

    def __init__(self, shell):
        super(SasxMagics, self).__init__(shell)

    @cell_magic
    def sasx(self, line_param, cell):

        cell_parsed = sasx_parse(cell, self)

        if cell_parsed['status']==0:
            print(cell_parsed['message'])
            return

        #Generate python code
        str_code = ""
        str_code = str_code + sasx_preloop(cell_parsed)
        str_code = str_code + sasx_loop(cell_parsed)
        str_code = str_code + sasx_postloop(cell_parsed)
        
        #Execute the code
        ns = {}
        print("-----")
        print(str_code)
        print("-----")
        exec str_code in self.shell.user_ns, ns

# Register
ip = get_ipython() 
ip.register_magics(SasxMagics)



def sasx_preloop(result):

	#Create series for variables modified in python code (and kept)

	df_in = result['df_in']
	output_variables = result['output_variables']

	str_code=""
	if len(result['python_blocks'])==1:

		for var in [var for var in output_variables]:
			str_code = str_code + "series_" + var + "  = pd.Series(index=" + df_in + ".index,name='" + var + "')\n"
	
	elif len(result['python_blocks'])>1:

		for var in [var for var in output_variables]:
			str_code = str_code + "series_" + var + "  = pd.Series(name='" + var + "')\n"

		str_code = str_code + "series_sasx_index  = pd.Series(name='sasx_index')\n"
	
	return str_code


def sasx_loop(result):

	df_in = result['df_in']
	input_variables = result['input_variables']
	output_variables = result['output_variables']
	indent = " " * result['first_line_indent']
	str_code = "i = 0\n"
	str_code = str_code + "for c in " + df_in + ".itertuples():\n"
	#Variables already existing in the DataFrame
	for var in [var for var in input_variables]:
		str_code = str_code +  indent + var + " = c." + var + "\n"

	
	#variables modified in python code (and kept)
	if len(result['python_blocks'])==1:

		python_lines = result['python_blocks'][0]
		str_code = str_code + '\n'.join(python_lines) + "\n"
		for var in [var for var in output_variables]:
			str_code = str_code + indent + "series_" + var + "[c.Index] = " + var + "\n"  
    
	elif len(result['python_blocks'])>1:

		for python_lines in result['python_blocks']:
			indent = " " * result['indents'].pop(0)
			str_code = str_code + indent + "#---start of python block---\n"
			str_code = str_code + '\n'.join(python_lines) + "\n"	
			str_code = str_code + indent + "#---end of python block---\n"
			for var in [var for var in output_variables]:
				str_code = str_code + indent + "series_" + var + ".loc[i] =" + var + "\n"  
			str_code = str_code + indent + "series_sasx_index.loc[i] = c.Index\n"  
			str_code = str_code +  indent + "i = i + 1\n"
	    
	return str_code
 
def sasx_postloop(result):

	df_in = result['df_in']
	df_out = result['df_out']
	output_variables = result['output_variables']
	drop = result['drop']
	keep = result['keep']

	str_code = ""

	#If ouput keyword
	if len(result['python_blocks'])>1:
		str_code = str_code + "df_sasx_tmp = pd.DataFrame()\n"
		for var in [var for var in output_variables]:
			str_code = str_code + "df_sasx_tmp['" + var + "'] = series_" + var + "\n"  
		str_code = str_code + "df_sasx_tmp['sasx_index'] = series_sasx_index\n"
		str_code = str_code + df_in + " = df_sasx_tmp.merge(" + df_in + ", how='left', left_on='sasx_index', right_index=True)\n"  

	#Update target DataFrame
	if df_out <> df_in:
		str_code = str_code + df_out + " = " + df_in + ".copy()\n"
	
	for var in [var for var in output_variables]:
		str_code = str_code + df_out + "['" + var + "'] = series_" + var + "\n"  

	#Drop & keep
	if (len(drop))>0:
		str_code = str_code + df_out + ".drop(['" + "','".join(drop) + "'], axis=1, inplace=True)\n"
	elif (len(keep))>0:
		str_code = str_code + df_out + "=" + df_out + "[['" + "','".join(keep) + "']]\n"

	#Add df_out DataFrame to globals()
	str_code = str_code + "globals()['" +  df_out + "'] = " + df_out + "\n"
	return str_code


import ast
import re
import pandas

#Check if list of names are valid python names
def invalid_names(names):
    for name in names:
        try:
            root=ast.parse(name)
        except SyntaxError:
            return name
        if not isinstance(root.body[0], ast.Expr):
            return name
        if not isinstance(root.body[0].value, ast.Name):
            return name
    #names are all valid names!
    return None

def sasx_parse(str_code, self):
    
    result = {}
    result['status'] = 0 # 1: OK; 0:Error
    result['message'] = "" # Error message
    result['df_out'] = None
    result['df_in'] = None
    result['drop'] = []
    result['keep'] = []
    result['first_line_indent'] = None
    result['python_blocks'] = []
    result['indents'] = []
    result['input_variables'] = [] #variables from input DataFrame used in the code
    result['output_variables'] = [] #variables modified in code (kept or not dropped)

    python_block = []

    #Read all lines in cell
    lines = str_code.split('\n')
    i = 0
    
    for line in lines:
        
        i = i + 1
        sline = line.strip()            
        
        #Keyword DATA
        if i == 1:
            if sline.upper().startswith('DATA ') and sline.endswith(':'):
                words = [x for x in re.split(" +",sline[:-1]) if x]
                if len(words) == 2 :
                    if invalid_names([words[1]]) is not None:
                        result['message'] = "SASX syntax error on line 1 : '" + invalid_names([words[1]]) + "' is not a valid name" 
                        return result
                    # words[1] is a valid name!
                    result['df_out'] = words[1]
                elif len(words) ==  1:
                    result['message'] = "SASX syntax error on line 1 : output name missing"
                    return result
                else:
                    result['message'] = "SASX syntax error on line 1 : only one output name expected"
                    return result   
            elif sline.upper().startswith('DATA '):
                result['message'] = "SASX syntax error on line 1 : line should end with ':'"
                return result
            else:
                result['message'] = "SASX syntax error on line 1 : keyword DATA expected"
                return result
            
        elif sline.upper().startswith('DATA '):
            result['message'] = "SASX syntax error on line " + str(i) + " : keyword DATA only allowed on first line"
            return result
        
        #Keyword SET
        elif sline.upper().startswith('SET '):
            words = [x for x in re.split(" +",sline) if x]
            if len(words) == 2 :
                if invalid_names([words[1]]) is not None:
                    result['message'] = "SASX syntax error on line " + str(i) + " : '" + invalid_names([words[1]]) + "' is not a valid name"
                    return result
                # words[1] is a valid name!
                if not isinstance(self.shell.user_ns.get(words[1]), pandas.core.frame.DataFrame):
                    result['message'] = "SASX syntax error on line " + str(i) + " : '" + words[1] + "' is not a valid DataFrame"
                    return result
                result['df_in'] = words[1]
            elif len(words) ==  1:
                result['message'] = "SASX syntax error on line " + str(i) + " : input name missing"
                return result    
            else:
                result['message'] = "SASX syntax error on line " + str(i) + " : only one input name expected"
                return result
            
        #Keyword DROP
        elif sline.upper().startswith('DROP '):
            
            #if keyword KEEP before
            if len(result['keep']) > 0:
                result['message'] = "SASX syntax error on line " + str(i) + " : keywords DROP and KEEP in the same data step"
                return result
            
            words = [x for x in re.split(" +",sline) if x]
            if len(words) ==  1:
                result['message'] = "SASX syntax error on line " + str(i) + " : variable name missing"
                return result
            if invalid_names(words[1:]) is not None:
                result['message'] = "SASX syntax error on line " + str(i) + " : '" + invalid_names([words[1]]) + "' is not a valid name"
                return result
            
            # words are all valid names!
            result['drop'] = words[1:]
            

        #Keyword KEEP
        elif sline.upper().startswith('KEEP '):
            
            #if keyword DROP before
            if len(result['drop']) > 0:
                result['message'] = "SASX syntax error on line " + str(i) + " : keywords DROP and KEEP in the same data step"
                return result 
            
            words = [x for x in re.split(" +",sline) if x]
            if len(words) ==  1:
                result['message'] = "SASX syntax error on line " + str(i) + " : variable name missing"
                return result
            if invalid_names(words[1:]) is not None:
                result['message'] = "SASX syntax error on line " + str(i) + " : '" + invalid_names([words[1]]) + "' is not a valid name"
                return result
            
            # words are all valid names!
            result['keep'] = words[1:]
            
        #Keyword OUTPUT
        elif sline.upper()=='OUTPUT':
            indent = len(line) - len(line.lstrip())
            #New block
            if len(python_block)>0:
                result['python_blocks'].append(python_block)
                result['indents'].append(indent)
            python_block = []


        #Python line (no SASX keyword)
        else:
            if (len(sline) > 0) :

                #Get indent if first python line
                if (len(python_block)==0) & (len(result['python_blocks'])==0):
                    indent = len(line) - len(line.lstrip())
                    result['first_line_indent'] = indent

                #Append line to current python block
                python_block.append(line)


    #Append last block to python_blocks only if no OUTPUT keyword before
    if (len(python_block)>0):
        if (len(result['python_blocks'])==0):
            result['python_blocks'].append(python_block)
        else:
            result['message'] = "SASX syntax error on line " + str(i) + " : keyword OUTPUT missing"
            return result
             
    #Parsing finished - doing some extra checks

    #At least 1 python line for correct syntax
    if len(result['python_blocks'])==0:
        result['message'] = "Error: no python line in the data step"
        return result

    #If df_in missing: df_out is used as the input DataFrame
    if (result['df_in']) == None:
        #Before assigning check that df_out is a DataFrame
        if not isinstance(self.shell.user_ns.get(result['df_out']), pandas.core.frame.DataFrame):
            result['message'] = "SASX syntax error on line 1 : '" + result['df_out'] + "' is not a valid DataFrame"
            return result
        result['df_in'] = result['df_out']

    #Parsing finished - get variables from input DataFrame and from the code

    #Get all variables from input Dataframe
    df_in_variables = self.shell.user_ns.get(result['df_in']).columns.values

    #Parse python lines
    python_lines = [item for sublist in result['python_blocks'] for item in sublist]
    root = ast.parse("def fake():\n" + "\n".join(python_lines))
    code_modified_variables = sorted({node.id for node in ast.walk(root) if isinstance(node, ast.Name) if isinstance(node.ctx, ast.Store)})
    code_variables = sorted({node.id for node in ast.walk(root) if isinstance(node, ast.Name) })

    #variables from input DataFrame used in the code
    result['input_variables'] = [var for var in code_variables if var in df_in_variables]

    #variables modified in code (not dropped or kept)
    if (len(result['keep'])==0):
        result['output_variables'] = [var for var in code_modified_variables if var not in result['drop'] ]
    elif (len(result['drop'])==0):
        result['output_variables'] = [var for var in code_modified_variables if var in result['keep'] ]

    #Parsing finished without errors
    result['status'] = 1
    return result




"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

version = '0.2'

install_requires = [
	'ipython',
	'pandas',
]

long_description = """
Data manipulation in Python for SAS users

SASX
----

Data manipulation in Python for SAS users, with the %%sasx magic command.

SASX (Simple dAta SyntaX) has the best of both worlds:
- Classic python syntax with access to python, numpy, pandas (like Python)
- A few extra keywords to allow easy row-by-row operations (like SAS)

Install the lastest release with:
pip install sasx

More info:  https://github.com/python-sasx/sasx
"""

setup(
    name='sasx',
    version=version,
    description='Data manipulation in Python for SAS users',
    long_description=long_description,
    url='https://github.com/python-sasx/sasx',
    author='python-sasx',
    author_email='python.sasx@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: IPython',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Database',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],

    keywords='data manipulation sas python pandas sasx ipython',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=install_requires,

)
